/* Copyright (c) 2012, Steffen Wittmeier, Michael Jäntsch
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * * Neither the name of the Technische Universitaet Muenchen nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "Muscle.h"
#include <string.h>

extern ::eu::er::fw::ecu::Muscle *mus;

namespace eu
{
namespace er
{
namespace fw
{
namespace ecu
{
Muscle::Muscle(u8 num, Heartbeat *hb) :
		mynum(num), hb(hb), initialized(false), t_s(0)
{
	if (0 == mynum)
	{
		canID = CAN_ID0;
	}
	else
	{
		canID = CAN_ID1;
	}

	states[com::robot::E_MS_Uninitialized].init = 0;
	states[com::robot::E_MS_Initialize].init = &Muscle::stateInitializeInit;
	states[com::robot::E_MS_Off].init = &Muscle::stateOffInit;
	states[com::robot::E_MS_On].init = &Muscle::stateOnInit;
	states[com::robot::E_MS_Deinitialize].init = 0;
	states[com::robot::E_MS_Error].init = 0;
	states[com::robot::E_MS_FatalError].init = 0;

	states[com::robot::E_MS_Uninitialized].execute = 0;
	states[com::robot::E_MS_Initialize].execute = 0;
	states[com::robot::E_MS_Deinitialize].execute = 0;
	states[com::robot::E_MS_Off].execute = &Muscle::stateOffExecute;
	states[com::robot::E_MS_On].execute = &Muscle::stateOnExecute;
	states[com::robot::E_MS_Error].execute = &Muscle::stateErrorExecute;
	states[com::robot::E_MS_FatalError].execute = &Muscle::stateErrorExecute;

	muscleControlMode = com::robot::VoltageControl;
	errno = com::robot::E_ERR_NONE;
	state = com::robot::E_MS_Uninitialized;
	lastState = com::robot::E_MS_NoState;
	refValue = 0.0f;

}

Muscle::~Muscle()
{
}

void Muscle::init(u8 * buffer)
{
	com::robot::MuscleInitData controllerParams;

	state = com::robot::E_MS_Off;

	com::robot::deserializeS32Array((s32 *) &controllerParams, 0,
			sizeof(com::robot::MuscleInitData), buffer);

	//voltages are in millivolts
	maxPosVoltage = (s32) (::el::math::Fixed().deserializeFrom(
			controllerParams.maxposVoltage) * 1000);
	minPosVoltage = (s32) (::el::math::Fixed().deserializeFrom(
			controllerParams.minposVoltage) * 1000);
	maxNegVoltage = (s32) (::el::math::Fixed().deserializeFrom(
			controllerParams.maxnegVoltage) * 1000);
	minNegVoltage = (s32) (::el::math::Fixed().deserializeFrom(
			controllerParams.minnegVoltage) * 1000);

	pMotorPosition = (s32) (::el::math::Fixed().deserializeFrom(
			controllerParams.PPoti));
	epsPoti = (s32) (::el::math::Fixed().deserializeFrom(
			controllerParams.EpsPoti));

	encoderConversion.deserializeFrom(controllerParams.encoderConversion);

	maxMotorPosition = controllerParams.maxMotorPosition;
	minMotorPosition = controllerParams.minMotorPosition;

	maxForce = (s32) (::el::math::Fixed().deserializeFrom(
			controllerParams.maxForce) * 10);

	forceSign = controllerParams.forceInverted ? -1 : 1;
	forceOffset = (s32) (::el::math::Fixed().deserializeFrom(
			controllerParams.forceOffset) * 10);
	forceGain.deserializeFrom(controllerParams.forceGain);

	forceCompensation.deserializeFrom(controllerParams.forceCompensation);

	el::math::Fixed initPos;

	if (mynum == 0)
	{
		initPos = MOTOR_POS0;
	}
	else
	{
		initPos = MOTOR_POS1;
	}

	encoder.init((s32) (initPos / encoderConversion));

	initialized = true;

}

void Muscle::handleMsg(const u8 cmd, const u8 dlc, const u8 * data)
{
	switch (cmd)
	{
	case com::robot::MuscleOffMsg:
		if (com::robot::E_MS_On == state)
		{
			state = com::robot::E_MS_Off;
		}

		if (com::robot::E_MS_Off == state)
		{
			canBus.send(
					com::robot::CAN_MAKE_CAN_ID(canID,
							com::robot::AcknowledgeMsg));
		}
		break;
	case com::robot::ReferenceValueSetMsg:
		if (state == com::robot::E_MS_On || state == com::robot::E_MS_Off)
		{
			muscleControlMode = (com::robot::ControlVariable) data[0];

			refValue.deserializeFrom(com::robot::deserializeS32(&data[1]));
		}

		break;
	case com::robot::MuscleOnMsg:
		if (com::robot::E_MS_Off == state)
		{
			state = com::robot::E_MS_On;
		}

		if (com::robot::E_MS_On == state)
		{
			canBus.send(
					com::robot::CAN_MAKE_CAN_ID(canID,
							com::robot::AcknowledgeMsg));
		}
		break;
	case com::robot::ResetMsg:
		reboot = true;
		break;
	case com::robot::ProprioceptiveSensorDataRequestMsg:
	{
		if (state == com::robot::E_MS_On || state == com::robot::E_MS_Off)
		{
			CanTxMsg reply;

			reply.IDE = CAN_ID_STD;
			reply.RTR = CAN_RTR_DATA;
			reply.StdId = com::robot::CAN_MAKE_CAN_ID(canID,
					com::robot::MuscleProprioceptiveSensorDataResponseMsg);
			reply.DLC = 8;

			com::robot::serializeU16(getForce(), &reply.Data[0]);
			com::robot::serializeU16((s32) getMotorPosition(), &reply.Data[2]);
			com::robot::serializeS16(getCurrent(), &reply.Data[4]);
			com::robot::serializeS16(0, &reply.Data[6]);

			canBus.send(&reply);
		}
		break;
	}
	case com::robot::MuscleInitMsg:
	{
		if (state == com::robot::E_MS_Uninitialized
				|| state == com::robot::E_MS_Initialize)
		{
			const u8 numInitMsg = data[0];

			static u8 buffer[sizeof(com::robot::MuscleInitData)];

			if (0 == numInitMsg)
			{
				paramIndex = 0;
				state = com::robot::E_MS_Initialize;
			}
			else if (state != com::robot::E_MS_Initialize
					|| paramIndex != numInitMsg)
			{
				errno = com::robot::E_ERR_PROTOCOL;
				break;
			}

			//Security check: do not overwrite controller parameter memory boundaries
			const u16 arrayBound = min(sizeof(com::robot::MuscleInitData)
					- numInitMsg * 7, 7);

			if ((numInitMsg * 7 + arrayBound)
					<= sizeof(com::robot::MuscleInitData))
			{
				memcpy(buffer + numInitMsg * 7, data + 1, arrayBound);
			}

			paramIndex++;

			if (paramIndex >= com::robot::NUM_MUSCLE_INIT_MESSAGES)
			{
				canBus.send(
						com::robot::CAN_MAKE_CAN_ID(canID,
								com::robot::AcknowledgeMsg));
				init(buffer);
			}
		}
		break;
	}

	case com::robot::MuscleDeinitMsg:
	{
		if (state == com::robot::E_MS_Off || state == com::robot::E_MS_On)
		{
			state = com::robot::E_MS_Uninitialized;
		}
		break;
	}

	case com::robot::MuscleStatusRequestMsg:
		canBus.send(
				com::robot::CAN_MAKE_CAN_ID(canID,
						com::robot::MuscleStatusResponseMsg), 2, state, errno);
		//reset error code
		errno = com::robot::E_ERR_NONE;
		break;
	}

	if (errno > com::robot::E_ERRLEVEL_WARNING)
	{
		state = com::robot::E_MS_Error;
	}
}

s32 Muscle::getMotorPosition()
{
	return encoder.getAngPosition() * encoderConversion;
}

void Muscle::updateSensorData()
{
	sense.startConv();
}

void Muscle::updateModel()
{
	if (state != lastState)
	{
		if (0 != states[state].init)
		{
			(this->*(states[state].init))();
		}
		lastState = state;
	}

	//Here we can use the time for computation until ADC conversion is done.

	while (!sense.convFinished())
		;

	force = forceSign
			* (s32) (el::math::Fixed(sense.getUnamplifiedForce(mynum))
					* forceGain * 10) - forceOffset;

	current = sense.getCurrent(mynum);
}

void Muscle::execControl()
{
	if (0 != states[state].execute)
	{
		(this->*(states[state].execute))();
	}
}

void Muscle::heartbeatMissing()
{
	hb->stop();

	if (state == com::robot::E_MS_On || state == com::robot::E_MS_Off)
	{
		motor.off();
		state = com::robot::E_MS_Deinitialize;
	}
}

void Muscle::stateInitializeInit()
{
}

void Muscle::stateOffInit()
{
	motor.off();
}

void Muscle::stateOffExecute()
{
}

void Muscle::stateOnInit()
{
	motor.on();

}

void Muscle::stateOnExecute()
{

	s32 voltage = 0;

	switch (muscleControlMode)
	{
	case com::robot::VoltageControl:
	{
		voltage = refValue * 1000;
		break;
	}
	case com::robot::MotorPositionControl:
	{

		//TODO: implement the motor position controller here!

		//check voltage against limits
		if (voltage > 0)
		{
			if (voltage > maxPosVoltage)
			{
				voltage = maxPosVoltage;
			}
			else if (voltage < minPosVoltage)
			{
				voltage = minPosVoltage;
			}
		}
		else if (voltage < 0)
		{
			if (voltage < maxNegVoltage)
			{
				voltage = maxNegVoltage;
			}
			else if (voltage > minNegVoltage)
			{
				voltage = minNegVoltage;
			}
		}
		break;

	}
	}

	//if motor was turning in the different direction before, halt it for one cycle.
	//(neccessary because otherwise there may be high currents in the hbridge)
	s16 old_voltage = getVoltage();

	if ((old_voltage > 0) && (voltage < 0))
		voltage = 0;
	else if ((old_voltage < 0) && (voltage > 0))
		voltage = 0;

	//if motor position is out of range, stop immediatly, unless motor is moving towards range
	if ((getMotorPosition() > maxMotorPosition) && (voltage > 0))
	{
		//motor.off();
		voltage = 0;
	}

	if ((getMotorPosition() < minMotorPosition) && (voltage < 0))
	{
		//motor.off();
		voltage = 0;
	}

	if ((getForce() > maxForce) && (voltage > 0))
	{
		voltage = 0;
	}

	//set voltage to motor
	motor.setVoltage(voltage);
}

void Muscle::stateErrorExecute()
{
	motor.off();
}
}
}
}
}
