/* Copyright (c) 2012, Michael Jaentsch, Steffen Wittmeier
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * * Neither the name of the Technische Universitaet Muenchen nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _COM_PROTOCOL_PROTOCOL_H_
#define _COM_PROTOCOL_PROTOCOL_H_

#include "libmath/Types.h"

namespace eu
{
namespace er
{
namespace com
{
namespace robot
{
enum MessageType
{
	Undefined = 0x00,
	MuscleOffMsg = 0x01,
	MuscleErrorMsg = 0x02,
	AcknowledgeMsg = 0x04,
	ResetMsg = 0x05,
	ReferenceValueSetMsg = 0x08,
	MuscleOnMsg = 0x09,
	HeartBeatMsg = 0x0A,
	ProprioceptiveSensorDataRequestMsg = 0x10,
	MuscleProprioceptiveSensorDataResponseMsg = 0x11,
	JointProprioceptiveSensorDataResponseMsg = 0x12,
	MuscleInitMsg = 0x18,
	JointInitMsg = 0x19,
	MuscleStatusRequestMsg = 0x1A,
	MuscleStatusResponseMsg = 0x1B,
	MuscleDeinitMsg = 0x1C,
	FirmwareInfoRequestMsg = 0x1E,
	FirmwareInfoResponseMsg = 0x1F
};

enum MuscleState
{
	E_MS_Uninitialized,
	E_MS_Initialize,
	E_MS_Deinitialize,
	E_MS_Off,
	E_MS_On,
	E_MS_Error,
	E_MS_FatalError,
	///No valid state. Always has to be last in enumeration.
	E_MS_NoState
};

const long NumMuscleStates = 7;

enum ControlVariable
{
	VoltageControl, MotorPositionControl, CurrentControl, ForceControl
};

enum E_ErrorCode
{
	///No error.
	E_ERR_NONE,
	///The heartbeat message was not received in the expected interval.
	E_ERR_HEARTBEAT_MISSED,
	///Errors below this level do not cause a transition to the error state.
	E_ERRLEVEL_WARNING,
	///Potentiometer is out of valid measurement range (which is defined in state_on_execute).
	E_ERR_POTI_OUT_OF_RANGE,
	///H bridge has error flag set.
	E_ERR_HBRIDGE_FAULT,
	///Error in the sequence of CAN messages.
	E_ERR_PROTOCOL,
	///Assertion in the firmware failed.
	E_ERR_ASSERTION,
	///High error level, which will cause a transition to the error state.
	E_ERRLEVEL_SEVERE,
	///The system encoutered a CAN error which prohibits reinitialization of the CAN controller and transmission of messages.
	E_ERR_SEVERE_CAN_ERROR,
	///Highest existing error level; there is no communication with the host possible.
	E_ERRLEVEL_FATAL
};

///Is this error code an error?
#define IS_ERROR(err) (err > E_ERR_NONE)
///Is the error code a warning?
#define IS_ERRORLEVEL_WARNING(err) (err > E_ERR_NONE && err < E_ERRLEVEL_WARNING)
///Is the error code a severe error?
#define IS_ERRORLEVEL_SEVERE(err) (err > E_ERRLEVEL_WARNING)
///Is the error code a fatal error?
#define IS_ERRORLEVEL_FATAL(err) (err > E_ERRLEVEL_SEVERE)

struct MuscleInitData
{
	s32 rotationInversed;

	s32 maxposVoltage;

	s32 minposVoltage;

	s32 maxnegVoltage;

	s32 minnegVoltage;

	s32 maxMotorPosition;

	s32 minMotorPosition;

	s32 maxForce;

	s32 PCurrent;

	s32 ICurrent;

	s32 DCurrent;

	s32 PPoti;

	s32 EpsPoti;

	s32 encoderConversion;

	s32 forceInverted;

	s32 forceGain;

	s32 forceOffset;

	s32 force_v;

	s32 force_kT1;

	s32 force_kT2;

	s32 force_kT3;

	s32 forceCompensation;

	s32 isSlave;
}__attribute__((packed));

static const unsigned char NUM_MUSCLE_INIT_MESSAGES = (sizeof(MuscleInitData)
		+ 6) / 7;

static const unsigned short MUSCLE_INIT_LAST_MESSAGE_LENGTH =
		sizeof(MuscleInitData) % 7;

static const u8 CAN_NUM_CONTROLLER_ID_BITS = 6;

typedef u16 can_id_t;

static const can_id_t CAN_CONTROLLER_ID_MASK = ((1
		<< (CAN_NUM_CONTROLLER_ID_BITS)) - 1);

static const can_id_t CAN_IDENTIFIER_MASK = 0x7ff;

static const can_id_t CAN_COMMAND_MASK = CAN_IDENTIFIER_MASK
		& (~CAN_CONTROLLER_ID_MASK);

inline can_id_t CAN_GET_CONTROLLER_ID(can_id_t can_id)
{
	return can_id & CAN_CONTROLLER_ID_MASK;
}

inline can_id_t CAN_GET_COMMAND(can_id_t can_id)
{
	return (can_id & CAN_COMMAND_MASK) >> CAN_NUM_CONTROLLER_ID_BITS;
}

inline can_id_t CAN_MAKE_CAN_ID(can_id_t controller_id, can_id_t command)
{
	return controller_id | (command << CAN_NUM_CONTROLLER_ID_BITS);
}

static const can_id_t BROADCAST_CONTROLLER_ID = 0;

inline void serializeU16(const u16 data, u8 * buffer)
{
	buffer[0] = (data >> 8) & 0xff;
	buffer[1] = data & 0xff;
}

inline void serializeS16(const s16 data, u8 * buffer)
{
	buffer[0] = (data >> 8) & 0xff;
	buffer[1] = data & 0xff;
}

inline void serializeU32(const u32 data, u8 * buffer)
{
	buffer[0] = (data >> 24) & 0xff;
	buffer[1] = (data >> 16) & 0xff;
	buffer[2] = (data >> 8) & 0xff;
	buffer[3] = data & 0xff;
}

inline void serializeS32(const s32 data, u8 * buffer)
{
	buffer[0] = (data >> 24) & 0xff;
	buffer[1] = (data >> 16) & 0xff;
	buffer[2] = (data >> 8) & 0xff;
	buffer[3] = data & 0xff;
}

inline u16 deserializeU16(const u8 * buffer)
{
	return ((u16) buffer[0] << 8) | buffer[1];
}

inline s16 deserializeS16(const u8 * buffer)
{
	return ((s16) buffer[0] << 8) | buffer[1];
}

inline u32 deserializeU32(const u8 * buffer)
{
	return ((u32) buffer[0] << 24) | (buffer[1] << 16) | (buffer[2] << 8)
			| buffer[3];
}

inline s32 deserializeS32(const u8 * buffer)
{
	return ((s32) buffer[0] << 24) | (buffer[1] << 16) | (buffer[2] << 8)
			| buffer[3];
}

inline void serializeS32Array(const s32 * data, u16 start, u16 length,
		u8 * buffer)
{
	s32* s32Data = (s32 *) buffer;

	for (u16 i = 0; i < length / sizeof(s32); i++)
		serializeS32(data[i], buffer + i * sizeof(s32));
}

inline void deserializeS32Array(s32 *data, u16 start, u16 length,
		const u8 * buffer)
{
	s32* s32Data = (s32 *) buffer;

	for (u16 i = 0; i < length / sizeof(s32); i++)
		data[i] = deserializeS32((u8*) (s32Data + i));
}
}
}
} /* namespace er */
} /* namespace eu */

#endif
