/* Copyright (c) 2012, Steffen Wittmeier, Michael Jäntsch
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * * Neither the name of the Technische Universitaet Muenchen nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "SysTimer.h"
#include "Led.h"

namespace eu
{
namespace er
{
namespace fw
{
namespace ecu
{
namespace hal
{

vu8 SysTimer::timer_flag = 0;
volatile SysTimer::Time SysTimer::t;

u16 SysTimer::period = 0;

/**
 * @brief Stop timer execution.
 */
void SysTimer::stop()
{
	/* stop/clear the systick timer and disable interrupt */
	SysTick_CounterCmd(SysTick_Counter_Disable);
	SysTick_CounterCmd(SysTick_Counter_Clear);
	SysTick_ITConfig( DISABLE);
}

/**
 * @brief Start periodic timer execution.
 *
 * @param period in us Time period for which the timer will run before expiration.
 * 	The least significant digit of <code>period</code> is ignored, values
 * 	between 0 and 7281 are valid (larger values will produce an overflow).
 */
void SysTimer::start(u16 period)
{
	SysTimer::period = period;
	t.microSeconds = 0;
	t.seconds = 0;

	/* SysTick end of count event each period (us)
	 * SysTick frequency is (72MHz/8 = 9 MHz, default)
	 * max value for period in us = 0xffff / 9 = 7281 */
	SysTick_SetReload(period * 9);

	/* Enable SysTick interrupt */
	SysTick_ITConfig( ENABLE);

	/* Start SysTick */
	SysTick_CounterCmd(SysTick_Counter_Enable);
}

/**
 * @brief Reset control loop not realtime error flag and wait till the timer expires.
 */
void SysTimer::wait()
{

	/* clear flag to indicate that computation has finished */
	timer_flag = 0;

	/* wait for IRQ SysTickHandler to set timer flag */
	while (!timer_flag)
	{
	}
}

/**
 * @brief Get system timer value.
 * @return Timer value in us.
 */
u32 SysTimer::getValue()
{
	return SysTick_GetCounter() / 9;
}

}
}
}
}
}

extern "C"
{
void __attribute__((__interrupt__))
SysTickHandler(void)
{
	/* the timer should not restart before computation has finished */
	//TODO
	assert_param(timer_flag == 0);

	if (::eu::er::fw::ecu::hal::SysTimer::timer_flag)
	{
		::eu::er::fw::ecu::hal::Led led(0);
		led.on();
	}

	::eu::er::fw::ecu::hal::SysTimer::t.microSeconds += ::eu::er::fw::ecu::hal::SysTimer::period;

	if (::eu::er::fw::ecu::hal::SysTimer::t.microSeconds > 1000000)
	{
		::eu::er::fw::ecu::hal::SysTimer::t.microSeconds = 0;
		::eu::er::fw::ecu::hal::SysTimer::t.seconds += 1;
	}

	/* set the flag indicating the timer has run out */
	::eu::er::fw::ecu::hal::SysTimer::timer_flag = 1;
}
}

