/*
 Copyright (c) 2006 Henry Strickland & Ryan Seto

 Permission is hereby granted, free of charge, to any person obtaining a
 copy of this software and associated documentation files (the "Software"),
 to deal in the Software without restriction, including without limitation
 the rights to use, copy, modify, merge, publish, distribute, sublicense,
 and/or sell copies of the Software, and to permit persons to whom the
 Software is furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included
 in all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 OTHER DEALINGS IN THE SOFTWARE.

 (* http://www.opensource.org/licenses/mit-license.php *)
 */

#ifndef EL_MATH_FIXED_H_
#define EL_MATH_FIXED_H_

#include "Types.h"

/*
#ifdef LINUX
#include <iostream>
#endif*/

namespace el
{
  namespace math
  {

    class Fixed
    {

    private:

      s32 g; // the guts

      const static u8 BP = 14; // how many low bits are right of Binary Point
      const static u8 BP2 = BP * 2; // how many low bits are right of Binary Point
      const static u8 BPhalf = BP / 2; // how many low bits are right of Binary Point

      static double
      STEP()
      {
        return 1.0 / (1 << BP);
      } // smallest step we can represent

      // for private construction via guts
      enum FixedRaw
      {
        RAW
      };
      Fixed(FixedRaw, s32 guts) :
        g(guts)
      {
      }

    public:
      Fixed() :
        g(0)
      {
      }
      Fixed(const Fixed& a) :
        g(a.g)
      {
      }
      Fixed(const float a) :
        g(int(a / (float) STEP()))
      {
      }
      Fixed(const double a) :
        g(int(a / (double) STEP()))
      {
      }
      Fixed(const s32 a) :
        g(a << BP)
      {
      }

      s32
      serialize()
      {
        return g;
      }

      Fixed &
      deserializeFrom(s32 buf)
      {
        g = buf;
        return *this;
      }

      Fixed&
      operator =(const Fixed& a)
      {
        g = a.g;
        return *this;
      }
      Fixed&
      operator =(const float a)
      {
        g = Fixed(a).g;
        return *this;
      }
      Fixed&
      operator =(const double a)
      {
        g = Fixed(a).g;
        return *this;
      }
      Fixed&
      operator =(const s32 a)
      {
        g = Fixed(a).g;
        return *this;
      }

      operator float() const
      {
        return g * (float) STEP();
      }
      operator double() const
      {
        return g * (double) STEP();
      }
      operator s32() const
      {
        return g >> BP;
      }

      Fixed
      operator +() const
      {
        return Fixed(RAW, g);
      }
      Fixed
      operator -() const
      {
        return Fixed(RAW, -g);
      }

      Fixed
      operator +(const Fixed& a) const
      {
        return Fixed(RAW, g + a.g);
      }
      Fixed
      operator -(const Fixed& a) const
      {
        return Fixed(RAW, g - a.g);
      }
#if 1
      // more acurate, using long long
      Fixed
      operator *(const Fixed& a) const
      {
        return Fixed(RAW, (s32)(((s64) g * a.g) >> BP));
      }
#else
      // faster, but with only half as many bits right of binary point
      Fixed operator *(const Fixed& a) const
      { return Fixed(RAW, (g>>BPhalf) * (a.g>>BPhalf) );}
#endif
      Fixed
      operator /(const Fixed& a) const
      {
        return Fixed(RAW, s32((((s64) g << BP2) / (s64)(a.g)) >> BP));
      }

      Fixed
      operator +(const float a) const
      {
        return Fixed(RAW, g + Fixed(a).g);
      }
      Fixed
      operator -(const float a) const
      {
        return Fixed(RAW, g - Fixed(a).g);
      }
      Fixed
      operator *(const float a) const
      {
        return Fixed(RAW, (g >> BPhalf) * (Fixed(a).g >> BPhalf));
      }
      Fixed
      operator /(const float a) const
      {
        return Fixed(RAW, int((((s64) g << BP2) / (s64)(Fixed(a).g)) >> BP));
      }

      Fixed
      operator +(const double a) const
      {
        return Fixed(RAW, g + Fixed(a).g);
      }
      Fixed
      operator -(const double a) const
      {
        return Fixed(RAW, g - Fixed(a).g);
      }
      Fixed
      operator *(const double a) const
      {
        return Fixed(RAW, (g >> BPhalf) * (Fixed(a).g >> BPhalf));
      }
      Fixed
      operator /(const double a) const
      {
        return Fixed(RAW, int((((s64) g << BP2) / (s64)(Fixed(a).g)) >> BP));
      }

      Fixed
      operator +(const s32 a) const
      {
        return Fixed(RAW, g + (a << BP));
      }
      Fixed
      operator -(const s32 a) const
      {
        return Fixed(RAW, g - (a << BP));
      }
      Fixed
      operator *(const s32 a) const
      {
        return Fixed(RAW, g * a);
      }
      Fixed
      operator /(const s32 a) const
      {
        return Fixed(RAW, g / a);
      }

      Fixed&
      operator +=(const Fixed &a)
      {
        return *this = *this + a;
        return *this;
      }
      Fixed&
      operator -=(const Fixed &a)
      {
        return *this = *this - a;
        return *this;
      }
      Fixed&
      operator *=(const Fixed &a)
      {
        return *this = *this * a;
        return *this;
      }
      Fixed&
      operator /=(const Fixed &a)
      {
        return *this = *this / a;
        return *this;
      }

      Fixed&
      operator +=(const s32 a)
      {
        return *this = *this + (Fixed) a;
        return *this;
      }
      Fixed&
      operator -=(const s32 a)
      {
        return *this = *this - (Fixed) a;
        return *this;
      }
      Fixed&
      operator *=(const s32 a)
      {
        return *this = *this * (Fixed) a;
        return *this;
      }
      Fixed&
      operator /=(const s32 a)
      {
        return *this = *this / (Fixed) a;
        return *this;
      }

      Fixed&
      operator +=(const float a)
      {
        return *this = *this + a;
        return *this;
      }
      Fixed&
      operator -=(const float a)
      {
        return *this = *this - a;
        return *this;
      }
      Fixed&
      operator *=(const float a)
      {
        return *this = *this * a;
        return *this;
      }
      Fixed&
      operator /=(const float a)
      {
        return *this = *this / a;
        return *this;
      }

      Fixed&
      operator +=(const double a)
      {
        return *this = *this + a;
        return *this;
      }
      Fixed&
      operator -=(const double a)
      {
        return *this = *this - a;
        return *this;
      }
      Fixed&
      operator *=(const double a)
      {
        return *this = *this * a;
        return *this;
      }
      Fixed&
      operator /=(const double a)
      {
        return *this = *this / a;
        return *this;
      }

      bool
      operator ==(const Fixed& a) const
      {
        return g == a.g;
      }
      bool
      operator !=(const Fixed& a) const
      {
        return g != a.g;
      }
      bool
      operator <=(const Fixed& a) const
      {
        return g <= a.g;
      }
      bool
      operator >=(const Fixed& a) const
      {
        return g >= a.g;
      }
      bool
      operator <(const Fixed& a) const
      {
        return g < a.g;
      }
      bool
      operator >(const Fixed& a) const
      {
        return g > a.g;
      }

      bool
      operator ==(const float a) const
      {
        return g == Fixed(a).g;
      }
      bool
      operator !=(const float a) const
      {
        return g != Fixed(a).g;
      }
      bool
      operator <=(const float a) const
      {
        return g <= Fixed(a).g;
      }
      bool
      operator >=(const float a) const
      {
        return g >= Fixed(a).g;
      }
      bool
      operator <(const float a) const
      {
        return g < Fixed(a).g;
      }
      bool
      operator >(const float a) const
      {
        return g > Fixed(a).g;
      }

      bool
      operator ==(const double a) const
      {
        return g == Fixed(a).g;
      }
      bool
      operator !=(const double a) const
      {
        return g != Fixed(a).g;
      }
      bool
      operator <=(const double a) const
      {
        return g <= Fixed(a).g;
      }
      bool
      operator >=(const double a) const
      {
        return g >= Fixed(a).g;
      }
      bool
      operator <(const double a) const
      {
        return g < Fixed(a).g;
      }
      bool
      operator >(const double a) const
      {
        return g > Fixed(a).g;
      }

      bool
      operator ==(const s32 a) const
      {
        return g == Fixed(a).g;
      }
      bool
      operator !=(const s32 a) const
      {
        return g != Fixed(a).g;
      }
      bool
      operator <=(const s32 a) const
      {
        return g <= Fixed(a).g;
      }
      bool
      operator >=(const s32 a) const
      {
        return g >= Fixed(a).g;
      }
      bool
      operator <(const s32 a) const
      {
        return g < Fixed(a).g;
      }
      bool
      operator >(const s32 a) const
      {
        return g > Fixed(a).g;
      }

#ifdef LINUX
      friend std::ostream&
      operator<<(std::ostream& os, const Fixed a)
      {
        os << a.g * (float) STEP();

        return os;
      }
#endif
    };

    inline Fixed
    operator +(const s32 a, const Fixed& b)
    {
      return Fixed(a) + b;
    }
    inline Fixed
    operator -(const s32 a, const Fixed& b)
    {
      return Fixed(a) - b;
    }
    inline Fixed
    operator *(const s32 a, const Fixed& b)
    {
      return Fixed(a) * b;
    }
    inline Fixed
    operator /(const s32 a, const Fixed& b)
    {
      return Fixed(a) / b;
    }

    inline Fixed
    operator +(const float a, const Fixed& b)
    {
      return Fixed(a) + b;
    }
    inline Fixed
    operator -(const float a, const Fixed& b)
    {
      return Fixed(a) - b;
    }
    inline Fixed
    operator *(const float a, const Fixed& b)
    {
      return Fixed(a) * b;
    }
    inline Fixed
    operator /(const float a, const Fixed& b)
    {
      return Fixed(a) / b;
    }

    inline bool
    operator ==(const float a, const Fixed& b)
    {
      return Fixed(a) == b;
    }
    inline bool
    operator !=(const float a, const Fixed& b)
    {
      return Fixed(a) != b;
    }
    inline bool
    operator <=(const float a, const Fixed& b)
    {
      return Fixed(a) <= b;
    }
    inline bool
    operator >=(const float a, const Fixed& b)
    {
      return Fixed(a) >= b;
    }
    inline bool
    operator <(const float a, const Fixed& b)
    {
      return Fixed(a) < b;
    }
    inline bool
    operator >(const float a, const Fixed& b)
    {
      return Fixed(a) > b;
    }

    inline Fixed
    operator +(const double a, const Fixed& b)
    {
      return Fixed(a) + b;
    }
    inline Fixed
    operator -(const double a, const Fixed& b)
    {
      return Fixed(a) - b;
    }
    inline Fixed
    operator *(const double a, const Fixed& b)
    {
      return Fixed(a) * b;
    }
    inline Fixed
    operator /(const double a, const Fixed& b)
    {
      return Fixed(a) / b;
    }

    inline bool
    operator ==(const double a, const Fixed& b)
    {
      return Fixed(a) == b;
    }
    inline bool
    operator !=(const double a, const Fixed& b)
    {
      return Fixed(a) != b;
    }
    inline bool
    operator <=(const double a, const Fixed& b)
    {
      return Fixed(a) <= b;
    }
    inline bool
    operator >=(const double a, const Fixed& b)
    {
      return Fixed(a) >= b;
    }
    inline bool
    operator <(const double a, const Fixed& b)
    {
      return Fixed(a) < b;
    }
    inline bool
    operator >(const double a, const Fixed& b)
    {
      return Fixed(a) > b;
    }

    inline s32&
    operator +=(s32& a, const Fixed& b)
    {
      a = (Fixed) a + b;
      return a;
    }
    inline s32&
    operator -=(s32& a, const Fixed& b)
    {
      a = (Fixed) a - b;
      return a;
    }
    inline s32&
    operator *=(s32& a, const Fixed& b)
    {
      a = (Fixed) a * b;
      return a;
    }
    inline s32&
    operator /=(s32& a, const Fixed& b)
    {
      a = (Fixed) a / b;
      return a;
    }

    inline float&
    operator +=(float& a, const Fixed& b)
    {
      a = a + b;
      return a;
    }
    inline float&
    operator -=(float& a, const Fixed& b)
    {
      a = a - b;
      return a;
    }
    inline float&
    operator *=(float& a, const Fixed& b)
    {
      a = a * b;
      return a;
    }
    inline float&
    operator /=(float& a, const Fixed& b)
    {
      a = a / b;
      return a;
    }

    inline double&
    operator +=(double& a, const Fixed& b)
    {
      a = a + b;
      return a;
    }
    inline double&
    operator -=(double& a, const Fixed& b)
    {
      a = a - b;
      return a;
    }
    inline double&
    operator *=(double& a, const Fixed& b)
    {
      a = a * b;
      return a;
    }
    inline double&
    operator /=(double& a, const Fixed& b)
    {
      a = a / b;
      return a;
    }

  }
}

#endif //_EL_MATH_FIXED_H_
