/* Copyright (c) 2012, Steffen Wittmeier, Michael Jäntsch
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * * Neither the name of the Technische Universitaet Muenchen nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "hal/libhal.h"
#include "com.robot/Protocol.h"
#include "Muscle.h"
#include "Heartbeat.h"

bool reboot;

using namespace ::eu::er;

fw::ecu::Muscle *mus;

fw::ecu::Heartbeat *hb;

void canMessageHandler(void)
{
	CanRxMsg RxMessage =
	{ 0, 0, 0, 0, 0,
	{ 0 }, 0 };

	CAN_Receive(CAN_FIFO0, &RxMessage);

	if (CAN_ID_STD == RxMessage.IDE)
	{
		const u32 muscleID = com::robot::CAN_GET_CONTROLLER_ID(RxMessage.StdId);

		const ::com::robot::can_id_t commandID = com::robot::CAN_GET_COMMAND(
				RxMessage.StdId);

		if (commandID == ::com::robot::HeartBeatMsg)
		{
			//handle heartbeat reset for both messages
			if (!(hb->isEnabled()))
			{
				hb->start();
			}
			hb->reset();
			return;
		}

		if (::com::robot::BROADCAST_CONTROLLER_ID == muscleID)
		{
			//handle broadcast message
			mus[0].handleMsg(commandID, RxMessage.DLC, RxMessage.Data);

		}
		else
		{
			if (mus[0].getCanID() == muscleID)
			{
				mus[0].handleMsg(commandID, RxMessage.DLC, RxMessage.Data);
			}
		}
	}
}

void heartbeat_missing(void)
{
	mus[0].heartbeatMissing();
}

int main(void)
{
	fw::ecu::hal::InternalClock::init();

	fw::ecu::hal::Led redLed(0);
	fw::ecu::hal::Led orangeLed(1);

	redLed.off();
	orangeLed.off();

	fw::ecu::Heartbeat h(200);
	hb = &h;

	fw::ecu::Muscle m(0, hb);
	mus = &m;

	//configure heartbeat missing handler
	hb->onHeartbeatMissing(heartbeat_missing);

	fw::ecu::hal::Can::setReceiveBroadcast(true);
	fw::ecu::hal::Can canBus;
	canBus.setCanInterruptHandler(canMessageHandler);

	reboot = false;

	u8 i;

	mus->setSampleTime(0.001);

	//start with system frequency of 2kHz, T=500us
	fw::ecu::hal::SysTimer::start(1000);

	while (true)
	{
		if (mus->getState() == ::com::robot::E_MS_Off)
		{
			orangeLed.on();
		}
		else
		{
			orangeLed.off();
		}

		if (mus->getState() == com::robot::E_MS_Deinitialize)
		{
			redLed.on();
		}
		else
		{
			//redLed.off();
		}

		mus->updateSensorData();
		mus->updateModel();
		mus->execControl();

		if (reboot)
		{
			NVIC_GenerateSystemReset();
		}
		fw::ecu::hal::SysTimer::wait();

	}
	return 1;
}

#ifdef DEBUG
void assert_failed(u8* file, u32 line)
{
	fw::ecu::hal::Led *myled = hal::Led::getLed(0);
	fw::ecu::hal::Motor *mymotor0 = hal::Motor::getMotor(0);
	fw::ecu::hal::Motor *mymotor1 = hal::Motor::getMotor(1);

	mymotor0->off();
	mymotor1->off();
	myled->off();

	while (1)
	{
	}

	UNUSED_PARAM(file);
	UNUSED_PARAM(line);
}
#endif

void HardFaultException(void)
{
	/* Go to infinite loop when Hard Fault exception occurs */
	while (1)
	{
	}
}

/*******************************************************************************
 * Function Name  : MemManageException
 * Description    : This function handles Memory Manage exception.
 * Input          : None
 * Output         : None
 * Return         : None
 *******************************************************************************/
void MemManageException(void)
{
	/* Go to infinite loop when Memory Manage exception occurs */
	while (1)
	{
	}
}

/*******************************************************************************
 * Function Name  : BusFaultException
 * Description    : This function handles Bus Fault exception.
 * Input          : None
 * Output         : None
 * Return         : None
 *******************************************************************************/
void BusFaultException(void)
{
	/* Go to infinite loop when Bus Fault exception occurs */
	while (1)
	{
	}
}

/*******************************************************************************
 * Function Name  : UsageFaultException
 * Description    : This function handles Usage Fault exception.
 * Input          : None
 * Output         : None
 * Return         : None
 *******************************************************************************/
void UsageFaultException(void)
{
	/* Go to infinite loop when Usage Fault exception occurs */
	while (1)
	{
	}
}
